"""
Benchmark comparison between acora search and re.findall()
"""

import re
import sys
import timeit
from time import time

from itertools import combinations
from functools import partial

try:
    from pyximport.pyxbuild import pyx_to_dll
except ImportError:
    pass
else:
    so_path = pyx_to_dll('acora/_acora.pyx')
    import sys, os.path
    sys.path.insert(0, os.path.dirname(so_path))

from acora import AcoraBuilder, BytesAcora, UnicodeAcora, PyAcora


def prepare_benchmark_data():
    s = ('bdfdaskdjfhaslkdhfsadhfklashdflabcasdabcdJAKHDBVDFLNFCBLSADHFCALKSJ'
        'jklhcnajskbhfasjhancfksjdfhbvaliuradefhzcbdegnashdgfbcjaabesdhgkfcnash'
        'fdkhbdegxcbgjsvdhabcabcfcgbnxahsdbgfbcakjsdhgnfcxsababcmdabe')
    s = s.lower() + s + s.upper()
    search_string = s * 1000

    all_keywords = [
        'ab', 'abc', 'abcd', 'abcabc', 'ababc', 'ABBBC', 'ABCABC',
        'bdfd', 'ade', 'abe', 'bdeg', 'fklash',
        'gnfcxsababcmdabe', 'SADHFCAL',
        'notthere', 'not-to-be-found', 'not-to-be-found-either',
        ]

    if sys.version_info[0] < 3:
        all_keywords = list(map(unicode, all_keywords))
        search_string = unicode(search_string)

    return search_string, all_keywords

def compare_search(s, filename, ignore_case, *keywords):
    t = time()
    builder = AcoraBuilder(keywords)
    py_acora = builder.build(ignore_case=ignore_case, acora=PyAcora)
    setup_pya = time() - t
    t = time()
    t = time()
    builder = AcoraBuilder(keywords)
    c_acora  = builder.build(ignore_case=ignore_case)
    setup_ca = time() - t
    if hasattr(keywords[0], 'encode'): # unicode in Py3?
        kw_regexp = '|'.join(keywords)
    else:
        kw_regexp = '|'.encode('ASCII').join(keywords)
    if ignore_case:
        regexp = re.compile(kw_regexp, re.I)
    else:
        regexp = re.compile(kw_regexp)
    setup_re = time() - t
    print("Case %ssensitive %s - setup times: PA: %.4f, CA: %.4f, RE: %.4f" % (
            ignore_case and 'in' or '',
            builder.for_unicode and 'unicode' or 'bytes',
            setup_pya, setup_ca, setup_re))

    timings = timeit.Timer(partial(py_acora.findall, s)).repeat(number=10)
    print("TIME(paS): %.3f" % min(timings))
    timings = timeit.Timer(partial(c_acora.findall, s)).repeat(number=10)
    print("TIME(caS): %.3f" % min(timings))
    if filename:
        timings = timeit.Timer(partial(py_acora.filefindall, filename)).repeat(number=10)
        print("TIME(paF): %.3f" % min(timings))
        timings = timeit.Timer(partial(c_acora.filefindall, filename)).repeat(number=10)
        print("TIME(caF): %.3f" % min(timings))
    timings = timeit.Timer(partial(regexp.findall, s)).repeat(number=10)
    print("TIME(reS): %.3f" % min(timings))

    return (c_acora.findall(s), py_acora.findall(s),
            filename and c_acora.filefindall(filename) or [],
            filename and py_acora.filefindall(filename) or [],
            regexp.findall(s))

def run_benchmark(search_string, all_keywords):
    search_string_lower = search_string.lower()
    bytes_search_string = search_string.encode('ASCII')
    bytes_search_string_lower = search_string_lower.encode('ASCII')

    import tempfile
    temp_text_file = tempfile.NamedTemporaryFile()
    temp_text_file.write(bytes_search_string)
    temp_text_file.flush()

    filename = temp_text_file.name

    for i in range(len(all_keywords),0,-1):
        for keywords in combinations(all_keywords, i):
            print('##Keywords(%d): %s' % (len(keywords), ' '.join(sorted(keywords))))
            keywords_lower = [ kw.lower() for kw in keywords ]

            results = compare_search(search_string, None, False, *keywords)
            assert_equal(results, results[0], search_string, keywords)
            assert_equal(results, results[1], search_string, keywords)

            results = compare_search(search_string, None, True, *keywords)
            assert_equal(results, results[0], search_string_lower, keywords_lower)
            assert_equal(results, results[1], search_string_lower, keywords_lower)

            keywords = [ keyword.encode('ASCII') for keyword in keywords ]

            results = compare_search(bytes_search_string, filename, False, *keywords)
            assert_equal(results, results[0], bytes_search_string, keywords)
            assert_equal(results, results[1], bytes_search_string, keywords)
            assert_equal(results, results[2], bytes_search_string, keywords)
            assert_equal(results, results[3], bytes_search_string, keywords)

            if sys.version_info[0] < 3:
                keywords_lower = [ keyword.encode('ASCII') for keyword in keywords_lower ]
                # case-insensitive search in byte strings is not supported in Py3
                results = compare_search(bytes_search_string, filename, True, *keywords)
                assert_equal(results, results[0], bytes_search_string_lower, keywords_lower)
                assert_equal(results, results[1], bytes_search_string_lower, keywords_lower)
                assert_equal(results, results[2], bytes_search_string_lower, keywords_lower)
                assert_equal(results, results[3], bytes_search_string_lower, keywords_lower)

def assert_equal(results, result, search_string, keywords):
    assert len(result) == sum(map(search_string.count, keywords)), \
        "EXPECTED: %d, got %s" % (
        sum(map(search_string.count, keywords)),
        list(map(len, results)))

if __name__ == '__main__':
    run_benchmark(*prepare_benchmark_data())



"""\
Acora - a multi-keyword search engine based on Aho-Corasick trees
and NFA2DFA powerset construction.

Usage::

    >>> from acora import AcoraBuilder

Collect some keywords::

    >>> builder = AcoraBuilder('ab', 'bc', 'de')
    >>> builder.add('a', 'b')

Generate the Acora search engine::

    >>> ac = builder.build()

Search a string for all occurrences::

    >>> ac.findall('abc')
    [('a', 0), ('ab', 0), ('b', 1), ('bc', 1)]
    >>> ac.findall('abde')
    [('a', 0), ('ab', 0), ('b', 1), ('de', 2)]
"""

try:
    unicode
except NameError:
    unicode = str

FILE_BUFFER_SIZE = 32 * 1024

class PyAcora(object):
    """A simple (and very slow) Python implementation of the Acora
    search engine.
    """
    def __init__(self, start_state, transitions):
        self.start_state = start_state.id
        self.transitions = dict([
                ((state.id, char), (target_state.id, target_state.matches))
                for ((state, char), target_state) in transitions.items() ])

    def finditer(self, s):
        """Iterate over all occurrences of any keyword in the string.

        Returns (keyword, offset) pairs.
        """
        state = self.start_state
        start_state = (state, [])
        next_state = self.transitions.get
        pos = 0
        for char in s:
            pos += 1
            state, matches = next_state((state,char), start_state)
            if matches:
                for match in matches:
                    yield (match, pos-len(match))

    def findall(self, s):
        """Find all occurrences of any keyword in the string.

        Returns a list of (keyword, offset) pairs.
        """
        return list(self.finditer(s))

    def filefind(self, f):
        """Iterate over all occurrences of any keyword in a file.

        Returns (keyword, offset) pairs.
        """
        opened = False
        if not hasattr(f, 'read'):
            f = open(f, 'rb')
            opened = True

        try:
            state = self.start_state
            start_state = (state, ())
            next_state = self.transitions.get
            pos = 0
            while 1:
                data = f.read(FILE_BUFFER_SIZE)
                if not data:
                    break
                for char in data:
                    pos += 1
                    state, matches = next_state((state,char), start_state)
                    if matches:
                        for match in matches:
                            yield (match, pos-len(match))
        finally:
            if opened:
                f.close()

    def filefindall(self, f):
        """Find all occurrences of any keyword in a file.

        Returns a list of (keyword, offset) pairs.
        """
        return list(self.filefind(f))

try:
    from acora._acora import UnicodeAcora, BytesAcora
except ImportError:
    # C module not there ...
    UnicodeAcora = BytesAcora = PyAcora

class AcoraBuilder(object):
    """The main builder class for an Acora search engine.

    Add keywords by calling ``.add(*keywords)`` or by passing them
    into the constructor. Then build the search engine by calling
    ``.build()``.
    """
    def __init__(self, *keywords):
        if len(keywords) == 1 and isinstance(keywords[0], (list, tuple)):
            keywords = keywords[0]
        self.for_unicode = None
        self.keywords = list(keywords)
        self.state_counter = 1
        self.tree = NfaState(0)
        self._insert_all(self.keywords)

    def add(self, *keywords):
        """Add more keywords to the search engine builder.

        Adding keywords does not impact previously built search
        engines.
        """
        self.keywords.extend(keywords)
        self._insert_all(keywords)

    def build(self, ignore_case=False, acora=None):
        """Build a search engine from the aggregated keywords.

        Builds a case insensitive search engine when passing
        ``ignore_case=True``, and a case sensitive engine otherwise.
        """
        if acora is None:
            if self.for_unicode:
                acora = UnicodeAcora
            else:
                acora = BytesAcora
        if self.for_unicode == False and ignore_case:
            import sys
            if sys.version_info[0] >= 3:
                raise ValueError("Case insensitive search is not supported for byte strings in Python 3")
        return acora(*nfa2dfa(self.tree, ignore_case))

    def _insert_all(self, keywords):
        for keyword in keywords:
            if self.for_unicode is None:
                self.for_unicode = isinstance(keyword, unicode)
            elif self.for_unicode != isinstance(keyword, unicode):
                raise TypeError(
                    "keywords must be either bytes or unicode, not mixed (got %s)" %
                    type(keyword))
            self.state_counter = insert_keyword(
                self.tree, keyword, self.state_counter)


# initial graph representation of the automaton

class NfaState(dict):
    """NFA state for the untransformed automaton.
    """
    def __new__(cls, state_id, *args, **kwargs):
        state = dict.__new__(cls, *args, **kwargs)
        state.id = state_id
        state.matches = []
        return state

    def __init__(self, state_id, *args, **kwargs):
        dict.__init__(self, *args, **kwargs)

    def __lt__(self, other):
        return self.id < other.id

    def __eq__(self, other):
        return self.id == other.id

    def __hash__(self):
        return self.id

    def __str__(self):
        return str(self.id)
    __repr__ = __str__

    def __copy__(self):
        state = NfaState(self.id, **self)
        state.matches[:] = self.matches
        return state

    def __deepcopy__(self, memo):
        state = NfaState(
            self.id,
            [ (char, state.__deepcopy__(None))
              for char, state in self.items() ])
        state.matches[:] = self.matches
        return state

def insert_keyword(tree, keyword, state_id):
    if not keyword:
        raise ValueError("cannot search for the empty string")
    for char in keyword:
        if char in tree:
            tree = tree[char]
        else:
            next_node = NfaState(state_id)
            state_id += 1
            tree[char] = next_node
            tree = next_node
    tree.matches = [keyword]
    return state_id

# NFA to DFA transformation

def nfa2dfa(tree, ignore_case):
    """Transform a keyword tree into a DFA using powerset construction.
    """
    def visit_all(tree, visitor):
        visitor(tree)
        for node in tree.values():
            visit_all(node, visitor)

    states = []
    visit_all(tree, states.append)
    next_state_id = len(states)

    # run through all states and collect all transitions, including
    # those from the start state (in which the NFA always stays)
    transitions = {}
    chars_by_state = {}
    new_eq_classes = set()
    for state in states:
        chars = chars_by_state[state] = set()
        for char, target in state.items():
            if ignore_case:
                char = char.lower()
            transitions[(state,char)] = set([target])
            chars.add(char)
        for char, target in tree.items():
            if ignore_case:
                char = char.lower()
            chars.add(char)
            key = (state,char)
            if key in transitions:
                transitions[key].add(target)
                new_eq_classes.add(key)
            else:
                transitions[key] = set([target])

    # create new states for newly found equivalence classes
    existing_eq_classes = {}
    eq_classes_by_state = {}
    while new_eq_classes:
        eq_classes = new_eq_classes
        new_eq_classes = set()
        for key in eq_classes:
            eq_states = transitions[key]
            if len(eq_states) < 2:
                continue
            eq_key = tuple(sorted([s.id for s in eq_states]))
            if eq_key in existing_eq_classes:
                transitions[key] = set([existing_eq_classes[eq_key]])
                continue

            # create a new joined state
            new_state = NfaState(next_state_id)

            matches = []
            for s in eq_states:
                matches.extend(s.matches)
            matches.sort(key=len, reverse=True)
            new_state.matches = matches

            eq_classes_by_state[new_state] = eq_states
            existing_eq_classes[eq_key] = new_state
            next_state_id += 1

            # redirect the original transition to the new node
            transitions[key] = set([new_state])

            # collect its transitions
            new_chars = chars_by_state[new_state] = set()
            for state in eq_states:
                chars = chars_by_state[state]
                new_chars.update(chars)
                for char in chars:
                    # resolve original states from equivalence class states
                    targets = set()
                    for target in transitions[(state,char)]:
                        if target in eq_classes_by_state:
                            targets.update(eq_classes_by_state[target])
                        else:
                            targets.add(target)
                    new_key = (new_state,char)
                    if new_key in transitions:
                        transitions[new_key].update(targets)
                    else:
                        transitions[new_key] = set(targets)
                    new_eq_classes.add(new_key)

    # rebuild transitions dict to point to exactly one state
    for key, state_set in transitions.items():
        assert len(state_set) == 1
        transitions[key] = tuple(state_set)[0]

    # duplicate the transitions for case insensitive parsing
    if ignore_case:
        for (state,char), target in list(transitions.items()):
            transitions[(state,char.upper())] = target

    # return start state and transitions
    return tree, transitions


### convenience functions

def search(s, *keywords):
    """Convenience function to search a string for keywords.
    """
    acora  = AcoraBuilder(keywords).build()
    return acora.findall(s)

def search_ignore_case(s, *keywords):
    """Convenience function to search a string for keywords.  Case
    insensitive version.
    """
    acora  = AcoraBuilder(keywords).build(ignore_case=True)
    return acora.findall(s)



#cython: embedsignature=True

"""A fast C implementation of the Acora search engine.

There are two main classes, UnicodeAcora and BytesAcora, that handle
byte data and unicode data respectively.
"""

__all__ = ['BytesAcora', 'UnicodeAcora']

cimport stdio
cimport python_exc
cimport python_mem
from python_ref cimport PyObject
from python_version cimport PY_MAJOR_VERSION
from python_unicode cimport Py_UNICODE, PyUnicode_AS_UNICODE, PyUnicode_GET_SIZE

cdef extern from * nogil:
     ctypedef Py_ssize_t ssize_t
     ssize_t read(int fd, void *buf, size_t count)

DEF MIN_TRANSITIONS_TO_BISECT = 20
DEF FILE_BUFFER_SIZE = 32 * 1024

ctypedef struct _AcoraUnicodeNodeStruct:
    Py_UNICODE* characters
    _AcoraUnicodeNodeStruct** targets
    PyObject** matches
    int char_count

ctypedef struct _AcoraBytesNodeStruct:
    char* characters
    _AcoraBytesNodeStruct** targets
    PyObject** matches
    int char_count

cdef _init_unicode_node(_AcoraUnicodeNodeStruct* c_node, state,
                        list state_transitions,
                        _AcoraUnicodeNodeStruct* all_nodes,
                        dict node_offsets, dict pyrefs):
    cdef size_t targets_mem_size, mem_size
    cdef Py_ssize_t i

    state_transitions.sort() # sort by characters
    characters, targets = list(zip(*state_transitions))
    characters = _intern(pyrefs, u''.join(characters))

    c_node.characters = PyUnicode_AS_UNICODE(characters)
    c_node.char_count = PyUnicode_GET_SIZE(characters)

    # use a single malloc for targets and match-string pointers
    mem_size = targets_mem_size = sizeof(_AcoraUnicodeNodeStruct**) * len(targets)
    if state.matches is not None and len(state.matches) > 0:
        mem_size += sizeof(PyObject*) * (len(state.matches) + 1) # NULL terminated
    c_node.targets = <_AcoraUnicodeNodeStruct**> python_mem.PyMem_Malloc(mem_size)
    if c_node.targets is NULL:
        python_exc.PyErr_NoMemory()

    for i, target in enumerate(targets):
        c_node.targets[i] = all_nodes + <size_t>node_offsets[target]

    if mem_size == targets_mem_size:
        c_node.matches = NULL
    else:
        c_node.matches = <PyObject**> (c_node.targets + len(targets))
        matches = _intern(pyrefs, tuple(state.matches))
        i = 0
        for match in matches:
            c_node.matches[i] = <PyObject*>match
            i += 1
        c_node.matches[i] = NULL

cdef _init_bytes_node(_AcoraBytesNodeStruct* c_node, state,
                      list state_transitions,
                      _AcoraBytesNodeStruct* all_nodes,
                      dict node_offsets, dict pyrefs):
    cdef size_t targets_mem_size, mem_size
    cdef Py_ssize_t i

    state_transitions.sort() # sort by characters
    characters, targets = list(zip(*state_transitions))
    if PY_MAJOR_VERSION >= 3:
        characters = bytes(characters) # items are integers, not byte strings
    else:
        characters = b''.join(characters)
    characters = _intern(pyrefs, characters)

    c_node.characters = characters
    c_node.char_count = len(characters)

    # use a single malloc for targets and match-string pointers
    mem_size = targets_mem_size = sizeof(_AcoraBytesNodeStruct**) * len(targets)
    if state.matches is not None and len(state.matches) > 0:
        mem_size += sizeof(PyObject*) * (len(state.matches) + 1) # NULL terminated
    c_node.targets = <_AcoraBytesNodeStruct**> python_mem.PyMem_Malloc(mem_size)
    if c_node.targets is NULL:
        python_exc.PyErr_NoMemory()

    for i, target in enumerate(targets):
        c_node.targets[i] = all_nodes + <size_t>node_offsets[target]

    if mem_size == targets_mem_size:
        c_node.matches = NULL
    else:
        c_node.matches = <PyObject**> (c_node.targets + len(targets))
        matches = _intern(pyrefs, tuple(state.matches))
        i = 0
        for match in matches:
            c_node.matches[i] = <PyObject*>match
            i += 1
        c_node.matches[i] = NULL

cdef inline _intern(dict d, obj):
    if obj in d:
        return d[obj]
    d[obj] = obj
    return obj


cdef dict group_transitions_by_state(dict transitions):
    # sort transitions by state ID (0 is start state) and transition character
    cdef list transition_list = sorted(transitions.items())
    transitions_by_state = {}
    for (state, character), target in transition_list:
        if state in transitions_by_state:
            transitions_by_state[state].append((character, target))
        else:
            transitions_by_state[state] = [(character, target)]
    return transitions_by_state


# unicode data handling

cdef class UnicodeAcora:
    """Acora search engine for unicode data.
    """
    cdef _AcoraUnicodeNodeStruct* start_node
    cdef Py_ssize_t node_count
    cdef bint has_large_nodes
    cdef tuple _pyrefs

    def __cinit__(self, start_state, dict transitions):
        cdef _AcoraUnicodeNodeStruct* c_nodes
        cdef Py_ssize_t i
        self.start_node = NULL
        cdef dict transitions_by_state = group_transitions_by_state(transitions)

        self.node_count = len(transitions_by_state)
        c_nodes = self.start_node = <_AcoraUnicodeNodeStruct*> python_mem.PyMem_Malloc(
            sizeof(_AcoraUnicodeNodeStruct) * self.node_count)
        if c_nodes is NULL:
            python_exc.PyMem_NoMemory()

        for i in range(self.node_count):
            # required by __dealloc__ in case of subsequent errors
            c_nodes[i].targets = NULL

        node_offsets = dict([ (state, i) for i,state in enumerate(transitions_by_state) ])
        pyrefs = {} # used to keep Python references alive (and intern them)
        for i, (state, state_transitions) in enumerate(transitions_by_state.iteritems()):
            if len(state_transitions) >= MIN_TRANSITIONS_TO_BISECT:
                self.has_large_nodes = True
            _init_unicode_node(&c_nodes[i], state, state_transitions,
                               c_nodes, node_offsets, pyrefs)

        self._pyrefs = tuple(pyrefs)

    def __dealloc__(self):
        cdef Py_ssize_t i
        if self.start_node is not NULL:
            for i in range(self.node_count):
                if self.start_node[i].targets is not NULL:
                    python_mem.PyMem_Free(self.start_node[i].targets)
            python_mem.PyMem_Free(self.start_node)

    cpdef finditer(self, unicode data):
        """Iterate over all occurrences of any keyword in the string.
        """
        if self.has_large_nodes:
            return _BisectUnicodeAcoraIter(self, data)
        else:
            return _UnicodeAcoraIter(self, data)

    def findall(self, unicode data):
        """Build a list of all occurrences of any keyword in the string.
        """
        return list(self.finditer(data))

cdef class _UnicodeAcoraIter:
    cdef _AcoraUnicodeNodeStruct* current_node
    cdef _AcoraUnicodeNodeStruct* start_node
    cdef Py_ssize_t match_index
    cdef unicode data
    cdef UnicodeAcora acora
    cdef Py_UNICODE* data_char
    cdef Py_UNICODE* data_end

    def __cinit__(self, UnicodeAcora acora, unicode data):
        assert acora.start_node is not NULL and acora.start_node.matches is NULL
        self.acora = acora
        self.start_node = self.current_node = acora.start_node
        self.match_index = 0
        self.data = data
        self.data_char = PyUnicode_AS_UNICODE(data)
        self.data_end = self.data_char + PyUnicode_GET_SIZE(data)

    def __iter__(self):
        return self

    def __next__(self):
        cdef Py_UNICODE* data_char = self.data_char
        cdef Py_UNICODE* data_end = self.data_end
        cdef Py_UNICODE* test_chars
        cdef Py_UNICODE current_char
        cdef int i, found = 0
        cdef _AcoraUnicodeNodeStruct* start_node = self.start_node
        cdef _AcoraUnicodeNodeStruct* current_node = self.current_node
        if current_node.matches is not NULL:
            if current_node.matches[self.match_index] is not NULL:
                return self._build_next_match()
            self.match_index = 0
        with nogil:
            while data_char < data_end:
                current_char = data_char[0]
                data_char += 1
                test_chars = current_node.characters
                if current_char < test_chars[0] \
                        or current_char > test_chars[current_node.char_count-1]:
                    current_node = start_node
                else:
                    for i in range(current_node.char_count):
                        if current_char == test_chars[i]:
                            current_node = current_node.targets[i]
                            break
                    else:
                        current_node = start_node
                    if current_node.matches is not NULL:
                        found = 1
                        break
        self.data_char = data_char
        self.current_node = current_node
        if found:
            return self._build_next_match()
        raise StopIteration

    cdef _build_next_match(self):
        match = <unicode> self.current_node.matches[self.match_index]
        self.match_index += 1
        return (match,
                <Py_ssize_t>(self.data_char - PyUnicode_AS_UNICODE(self.data)
                             ) - PyUnicode_GET_SIZE(match))

cdef class _BisectUnicodeAcoraIter(_UnicodeAcoraIter):
    def __next__(self):
        cdef Py_UNICODE* data_char = self.data_char
        cdef Py_UNICODE* data_end = self.data_end
        cdef Py_UNICODE* test_chars
        cdef Py_UNICODE current_char
        cdef int i, found = 0
        cdef _AcoraUnicodeNodeStruct* start_node = self.start_node
        cdef _AcoraUnicodeNodeStruct* current_node = self.current_node
        if current_node.matches is not NULL:
            if current_node.matches[self.match_index] is not NULL:
                return self._build_next_match()
            self.match_index = 0
        with nogil:
            while data_char < data_end:
                current_char = data_char[0]
                data_char += 1
                test_chars = current_node.characters
                if current_char < test_chars[0] \
                        or current_char > test_chars[current_node.char_count-1]:
                    current_node = start_node
                elif current_node.char_count >= MIN_TRANSITIONS_TO_BISECT:
                    i = _bisect(current_char, current_node.characters, current_node.char_count)
                    if i >= 0:
                        current_node = current_node.targets[i]
                    else:
                        current_node = start_node
                else:
                    for i in range(current_node.char_count):
                        if current_char == test_chars[i]:
                            current_node = current_node.targets[i]
                            break
                    else:
                        current_node = start_node
                if current_node.matches is not NULL:
                    found = 1
                    break
        self.data_char = data_char
        self.current_node = current_node
        if found:
            return self._build_next_match()
        raise StopIteration

cdef int _bisect(Py_UNICODE current_char, Py_UNICODE* characters, int char_count) nogil:
    cdef int start = 0, stop = char_count-1, mid = char_count // 2
    while start < stop:
        if current_char > characters[mid]:
            start = mid + 1
        elif current_char < characters[mid]:
            stop = mid - 1
        else:
            return mid
        mid = (start+stop) // 2
    if start == stop and current_char == characters[start]:
        return start
    return -1


# bytes data handling

cdef class BytesAcora:
    """Acora search engine for byte data.
    """
    cdef _AcoraBytesNodeStruct* start_node
    cdef Py_ssize_t node_count
    cdef tuple _pyrefs

    def __cinit__(self, start_state, dict transitions):
        cdef _AcoraBytesNodeStruct* c_nodes
        cdef Py_ssize_t i
        self.start_node = NULL
        cdef dict transitions_by_state = group_transitions_by_state(transitions)

        self.node_count = len(transitions_by_state)
        c_nodes = self.start_node = <_AcoraBytesNodeStruct*> python_mem.PyMem_Malloc(
            sizeof(_AcoraBytesNodeStruct) * self.node_count)
        if c_nodes is NULL:
            python_exc.PyMem_NoMemory()

        for i in range(self.node_count):
            # required by __dealloc__ in case of subsequent errors
            c_nodes[i].targets = NULL

        node_offsets = dict([ (state, i) for i,state in enumerate(transitions_by_state) ])
        pyrefs = {} # used to keep Python references alive (and intern them)
        for i, (state, state_transitions) in enumerate(transitions_by_state.iteritems()):
            _init_bytes_node(&c_nodes[i], state, state_transitions,
                             c_nodes, node_offsets, pyrefs)

        self._pyrefs = tuple(pyrefs)

    def __dealloc__(self):
        cdef Py_ssize_t i
        if self.start_node is not NULL:
            for i in range(self.node_count):
                if self.start_node[i].targets is not NULL:
                    python_mem.PyMem_Free(self.start_node[i].targets)
            python_mem.PyMem_Free(self.start_node)

    cpdef finditer(self, bytes data):
        return _BytesAcoraIter(self, data)

    def findall(self, bytes data):
        return list(self.finditer(data))

    def filefind(self, f):
        close_file = False
        if not hasattr(f, 'read'):
            f = open(f, 'rb')
            close_file = True
        return _FileAcoraIter(self, f, close_file)

    def filefindall(self, f):
        return list(self.filefind(f))

cdef class _BytesAcoraIter:
    cdef _AcoraBytesNodeStruct* current_node
    cdef _AcoraBytesNodeStruct* start_node
    cdef Py_ssize_t match_index
    cdef bytes data
    cdef BytesAcora acora
    cdef char* data_char
    cdef char* data_end
    cdef char* data_start

    def __cinit__(self, BytesAcora acora, bytes data):
        assert acora.start_node is not NULL and acora.start_node.matches is NULL
        self.acora = acora
        self.start_node = self.current_node = acora.start_node
        self.match_index = 0
        self.data_char = self.data_start = self.data = data
        self.data_end = self.data_char + len(data)

    def __iter__(self):
        return self

    def __next__(self):
        cdef char* data_char = self.data_char
        cdef char* data_end = self.data_end
        cdef char* test_chars
        cdef char current_char
        cdef int i, found = 0
        if self.current_node.matches is not NULL:
            if self.current_node.matches[self.match_index] is not NULL:
                return self._build_next_match()
            self.match_index = 0
        with nogil:
            found = _search_in_bytes(self.start_node, data_end,
                                     &self.data_char, &self.current_node)
        if found:
            return self._build_next_match()
        raise StopIteration

    cdef _build_next_match(self):
        match = <bytes> self.current_node.matches[self.match_index]
        self.match_index += 1
        return (match, <Py_ssize_t>(self.data_char - self.data_start) - len(match))

cdef int _search_in_bytes(_AcoraBytesNodeStruct* start_node,
                          char* data_end,
                          char** _data_char,
                          _AcoraBytesNodeStruct** _current_node) nogil:
    cdef char* data_char = _data_char[0]
    cdef _AcoraBytesNodeStruct* current_node = _current_node[0]
    cdef char* test_chars
    cdef char current_char
    cdef int i, found = 0

    while data_char < data_end:
        current_char = data_char[0]
        data_char += 1
        test_chars = current_node.characters
        if current_char < test_chars[0] \
                or current_char > test_chars[current_node.char_count-1]:
            current_node = start_node
        else:
            for i in range(current_node.char_count):
                if current_char == test_chars[i]:
                    current_node = current_node.targets[i]
                    break
            else:
                current_node = start_node
            if current_node.matches is not NULL:
                found = 1
                break
    _data_char[0] = data_char
    _current_node[0] = current_node
    return found

# file data handling

cdef class _FileAcoraIter:
    cdef _AcoraBytesNodeStruct* current_node
    cdef _AcoraBytesNodeStruct* start_node
    cdef Py_ssize_t match_index, read_size, buffer_offset_count
    cdef bytes buffer
    cdef char* c_buffer_pos
    cdef char* c_buffer_end
    cdef object f
    cdef bint close_file
    cdef int c_file
    cdef BytesAcora acora

    def __cinit__(self, BytesAcora acora, f, bint close=False, Py_ssize_t buffer_size=FILE_BUFFER_SIZE):
        assert acora.start_node is not NULL and acora.start_node.matches is NULL
        self.acora = acora
        self.start_node = self.current_node = acora.start_node
        self.match_index = 0
        self.buffer_offset_count = 0
        self.f = f
        self.close_file = close
        try:
            self.c_file = f.fileno()
        except:
            # maybe not a C file?
            self.c_file = -1
        self.read_size = buffer_size
        if self.c_file == -1:
            self.buffer = b''
        else:
            # use a statically allocated, fixed-size C buffer
            self.buffer = b'\0' * buffer_size
        self.c_buffer_pos = self.c_buffer_end = <char*> self.buffer

    def __iter__(self):
        return self

    def __next__(self):
        cdef bytes buffer
        cdef char* c_buffer
        cdef char* data_end
        cdef int error = 0, found = 0
        cdef Py_ssize_t buffer_size, bytes_read = 0
        if self.c_buffer_pos is NULL:
            raise StopIteration
        if self.current_node.matches is not NULL:
            if self.current_node.matches[self.match_index] is not NULL:
                return self._build_next_match()
            self.match_index = 0

        buffer_size = len(self.buffer)
        c_buffer = <char*> self.buffer
        if self.c_file != -1:
            with nogil:
                found = _find_next_match_in_cfile(
                    self.c_file, c_buffer, buffer_size, self.start_node,
                    &self.c_buffer_pos, &self.c_buffer_end,
                    &self.buffer_offset_count, &self.current_node, &error)
            if error:
                python_exc.PyErr_SetFromErrno(IOError)
        else:
            data_end = c_buffer + buffer_size
            while not found:
                if self.c_buffer_pos >= data_end:
                    self.buffer_offset_count += buffer_size
                    self.buffer = self.f.read(self.read_size)
                    buffer_size = len(self.buffer)
                    if buffer_size == 0:
                        self.c_buffer_pos = NULL
                        break
                    c_buffer = self.c_buffer_pos = <char*> self.buffer
                    data_end = c_buffer + buffer_size
                with nogil:
                    found = _search_in_bytes(
                        self.start_node, data_end,
                        &self.c_buffer_pos, &self.current_node)
        if self.c_buffer_pos is NULL:
            if self.close_file:
                self.f.close()
        elif found:
            return self._build_next_match()
        raise StopIteration

    cdef _build_next_match(self):
        match = <bytes> self.current_node.matches[self.match_index]
        self.match_index += 1
        return (match, self.buffer_offset_count + (
                self.c_buffer_pos - (<char*> self.buffer)) - len(match))


cdef int _find_next_match_in_cfile(int c_file, char* c_buffer, size_t buffer_size,
                                   _AcoraBytesNodeStruct* start_node,
                                   char** _buffer_pos, char** _buffer_end,
                                   Py_ssize_t* _buffer_offset_count,
                                   _AcoraBytesNodeStruct** _current_node,
                                   int* error) nogil:
    cdef char* buffer_pos = _buffer_pos[0]
    cdef char* buffer_end = _buffer_end[0]
    cdef char* data_end = c_buffer + buffer_size
    cdef Py_ssize_t buffer_offset_count = _buffer_offset_count[0]
    cdef _AcoraBytesNodeStruct* current_node = _current_node[0]
    cdef int found = 0
    cdef Py_ssize_t bytes_read

    while not found:
        if buffer_pos >= buffer_end:
            buffer_offset_count += buffer_end - c_buffer
            bytes_read = read(c_file, c_buffer, buffer_size)
            if bytes_read <= 0:
                if bytes_read < 0:
                    error[0] = 1
                buffer_pos = NULL
                break
            buffer_pos = c_buffer
            buffer_end = c_buffer + bytes_read

        found = _search_in_bytes(
            start_node, buffer_end, &buffer_pos, &current_node)

    _current_node[0] = current_node
    _buffer_offset_count[0] = buffer_offset_count
    _buffer_pos[0] = buffer_pos
    _buffer_end[0] = buffer_end
    return found



from distutils.core import setup
from distutils.extension import Extension

import sys
import os.path

try:
    from Cython.Distutils import build_ext
    cmdclass = {'build_ext': build_ext}
    source_ext = 'pyx'
except ImportError:
    build_ext = None
    cmdclass = {}
    source_ext = 'c'

version = "1.3"

setup(
    name = "acora",
    version = version,
    author="Stefan Behnel",
    author_email="stefan_ml@behnel.de",
    maintainer="Stefan Behnel",
    maintainer_email="stefan_ml@behnel.de",
    url="http://pypi.python.org/pypi/acora",
    download_url="http://pypi.python.org/packages/source/a/acora/acora-%s.tar.gz" % version,

    description="Fast multi-keyword search engine for text strings",

    long_description = open('README.txt').read(),

    classifiers = [
    'Intended Audience :: Developers',
    'Intended Audience :: Information Technology',
    'License :: OSI Approved :: BSD License',
    'Programming Language :: Cython',
    'Programming Language :: Python :: 2',
    'Programming Language :: Python :: 2.5',
    'Programming Language :: Python :: 2.6',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.0',
    'Programming Language :: Python :: 3.1',
    'Programming Language :: Python :: 3.2',
    'Operating System :: OS Independent',
    'Topic :: Text Processing',
    ],

    # extension setup

    cmdclass = {'build_ext': build_ext},
    ext_modules = [Extension("acora._acora",
                             ["acora/_acora."+source_ext])],
    packages = ['acora'],
)



"""
Simple test suite for acora.
"""

import acora

if acora.BytesAcora is acora.PyAcora or acora.UnicodeAcora is acora.PyAcora:
    print("WARNING: '_acora' C extension not imported, only testing Python implementation")

import re
import sys
import unittest
import codecs

# compat stuff ...

try:
    unicode
except NameError:
    unicode = str

try:
    bytes
except NameError:
    bytes = str

try:
    # Python 2.6+
    from io import StringIO as _StringIO, BytesIO as _BytesIO
except ImportError:
    # Python 2
    from StringIO import StringIO as _StringIO
    _BytesIO = _StringIO

def BytesIO(*args):
    if args and isinstance(args[0], unicode):
        args = (args[0].encode("UTF-8"),)
    return _BytesIO(*args)

def StringIO(*args):
    if args and isinstance(args[0], bytes):
        args = (args[0].decode("UTF-8"),)
    return _BytesIO(*args)

unicode_unescaper = codecs.lookup("unicode_escape")
def unescape_unicode(s):
    return unicode_unescaper.decode(s)[0]


def prepare_test_data():
    s = ('bdfdaskdjfhaslkdhfsadhfklashdflabcasdabcdJAKHDBVDFLNFCBLSADHFCALKSJ'
        'jklhcnajskbhfasjhancfksjdfhbvaliuradefhzcbdegnashdgfbcjaabesdhgkfcnash'
        'fdkhbdegxcbgjsvdhabcabcfcgbnxahsdbgfbcakjsdhgnfcxsababcmdabe')
    s = s.lower() + s + s.upper()
    search_string = s * 1000

    all_keywords = [
        'ab', 'abc', 'abcd', 'abcabc', 'ababc', 'ABBBC', 'ABCABC',
        'bdfd', 'ade', 'abe', 'bdeg', 'fklash',
        'gnfcxsababcmdabe', 'SADHFCAL',
        'notthere', 'not-to-be-found', 'not-to-be-found-either',
        ]

    if sys.version_info[0] < 3:
        all_keywords = list(map(unicode, all_keywords))
        search_string = unicode(search_string)

    return search_string, all_keywords


class AcoraTest(object):
    search_string, all_keywords = prepare_test_data()

    def _build(self, *keywords):
        keywords = list(map(self._swrap, keywords))
        return acora.AcoraBuilder(*keywords).build(acora=self.acora)

    def _build_ignore_case(self, *keywords):
        keywords = list(map(self._swrap, keywords))
        return acora.AcoraBuilder(*keywords).build(
            ignore_case=True, acora=self.acora)

    def _result(self, result):
        s = self._swrap
        return [ (s(k), pos) for k,pos in result ]

    # basic tests

    def test_finditer_single_keyword(self):
        s = self._swrap
        finditer = self._build('bc').finditer
        self.assertEquals(
            sorted(finditer(s('abcd'))),
            self._result([('bc', 1)]))

    def test_finditer_sequential(self):
        s = self._swrap
        finditer = self._build('a', 'b', 'c', 'd').finditer
        self.assertEquals(
            sorted(finditer(s('abcd'))),
            self._result([('a', 0), ('b', 1), ('c', 2), ('d', 3)]))

    def test_finditer_redundant(self):
        s = self._swrap
        finditer = self._build('a', 'b', 'A', 'B').finditer
        self.assertEquals(
            sorted(finditer(s('AaBb'))),
            self._result([('A', 0), ('B', 2), ('a', 1), ('b', 3)]))

    def test_finditer_overlap(self):
        s = self._swrap
        finditer = self._build('a', 'ab', 'abc', 'abcd').finditer
        self.assertEquals(
            sorted(finditer(s('abcd'))),
            self._result([('a', 0), ('ab', 0), ('abc', 0), ('abcd', 0)]))

    def test_finditer_reverse_overlap(self):
        s = self._swrap
        finditer = self._build('d', 'cd', 'bcd', 'abcd').finditer
        self.assertEquals(
            sorted(finditer(s('abcd'))),
            self._result([('abcd', 0), ('bcd', 1), ('cd', 2), ('d', 3)]))

    def test_deepcopy_builder(self):
        from copy import deepcopy
        s = self._swrap

        builder1 = acora.AcoraBuilder(*list(map(s, ['a', 'b', 'c'])))
        builder2 = deepcopy(builder1)
        builder2.add(s('ab'), s('bc'))
        
        finditer1 = builder1.build(acora=self.acora).finditer
        finditer2 = builder2.build(acora=self.acora).finditer

        self.assertEquals(
            sorted(finditer1(s('abcd'))),
            self._result([('a', 0), ('b', 1), ('c', 2)]))

        self.assertEquals(
            sorted(finditer2(s('abcd'))),
            self._result([('a', 0), ('ab', 0), ('b', 1), ('bc', 1), ('c', 2)]))
        

class UnicodeAcoraTest(unittest.TestCase, AcoraTest):
    # only unicode data tests
    from acora import UnicodeAcora as acora

    def _swrap(self, s):
        if not isinstance(s, unicode):
            s = s.decode('utf-8')
        return unescape_unicode(s)

    def test_finditer_line_endings(self):
        s = self._swrap
        finditer = self._build_ignore_case('a', 'b', 'c', 'd', '\r', '\n').finditer

        line = 0
        line_matches = []
        current_line_matches = []
        last_ending = None
        for kw, pos in finditer(s('Aa\r\nB\nbC\n\rcD\r\nd')):
            if kw in '\r\n':
                if last_ending == '\r' and kw == '\n':
                    continue
                line_matches.append(tuple(current_line_matches))
                del current_line_matches[:]
                last_ending = kw
                line += 1
            else:
                last_ending = None
                current_line_matches.append(kw)

        line_matches.append(tuple(current_line_matches))

        self.assertEquals(line, 5)
        self.assertEquals(
            line_matches,
            [('a', 'a'), ('b',), ('b', 'c'), (), ('c', 'd'), ('d',)])

    def test_finditer_single_keyword_unicode(self):
        s = self._swrap
        finditer = self._build(unicode("\\uF8D2")).finditer
        self.assertEquals(
            list(finditer(s(unicode("\\uF8D1\\uF8D2\\uF8D3")))),
            self._result([(unicode("\\uF8D2"), 1)]))

    def test_finditer_ignore_case(self):
        s = self._swrap
        finditer = self._build_ignore_case('a', 'b', 'c', 'd').finditer
        self.assertEquals(
            sorted(finditer(s('AaBbCcDd'))),
            self._result([('a', 0), ('a', 1), ('b', 2), ('b', 3),
                          ('c', 4), ('c', 5), ('d', 6), ('d', 7)]))

    def test_finditer_ignore_case_redundant(self):
        s = self._swrap
        finditer = self._build_ignore_case('a', 'b', 'A', 'B').finditer
        self.assertEquals(
            sorted(finditer(s('AaBb'))),
            self._result([('A', 0), ('A', 1), ('B', 2), ('B', 3),
                          ('a', 0), ('a', 1), ('b', 2), ('b', 3)]))


class BytesAcoraTest(unittest.TestCase, AcoraTest):
    # only byte data tests
    from acora import BytesAcora as acora

    simple_data = 'abc' + ('a'*100+'b'*100)*1000 + 'abcde'
    simple_kwds = ['abc'.encode('ASCII'),
                   'abcde'.encode('ASCII')]
    last_match_pos = len(simple_data) - 5
    expected_result = [(simple_kwds[0], 0),
                       (simple_kwds[0], last_match_pos),
                       (simple_kwds[1], last_match_pos)]

    def _swrap(self, s):
        if isinstance(s, unicode):
            s = s.encode('utf-8')
        return s

    def _search_in_file(self, ac, data):
        import tempfile
        tmp = tempfile.TemporaryFile()
        try:
            tmp.write(data.encode('ASCII'))
            tmp.seek(0)
            return list(ac.filefind(tmp))
        finally:
            tmp.close()

    def test_large_filelike_searching(self):
        filefind = self._build('SADHFCAL'.encode('ASCII'),
                               'bdeg'.encode('ASCII')).filefind
        data = BytesIO(self.search_string)
        result = list(filefind(data))
        self.assertEquals(len(result), 6000)

    def test_large_filelike_searching_check(self):
        ac = self._build(*self.simple_kwds)
        data = BytesIO(self.simple_data)
        result = list(ac.filefind(data))
        self.assertEquals(result, self.expected_result)

    def test_file_searching(self):
        ac = self._build([ kw.encode('ASCII')
                           for kw in ('a', 'b', 'ab', 'abc') ])
        result = self._search_in_file(ac, 'abbabc')
        self.assertEquals(len(result), 8)

    def test_large_file_searching(self):
        ac = self._build('SADHFCAL'.encode('ASCII'),
                         'bdeg'.encode('ASCII'))
        result = self._search_in_file(ac, self.search_string)
        self.assertEquals(len(result), 6000)

    def test_large_file_searching_check(self):
        ac = self._build(*self.simple_kwds)
        result = self._search_in_file(ac, self.simple_data)
        self.assertEquals(result, self.expected_result)


class PyAcoraTest(UnicodeAcoraTest, BytesAcoraTest):
    # both types of tests work here
    from acora import PyAcora as acora

    def _swrap(self, s):
        if isinstance(s, unicode):
            s = unescape_unicode(s)
        return s


def suite():
    import doctest
    suite = unittest.TestSuite([
            unittest.makeSuite(UnicodeAcoraTest),
            unittest.makeSuite(BytesAcoraTest),
            unittest.makeSuite(PyAcoraTest),
            doctest.DocTestSuite(),
            doctest.DocFileSuite('README.txt'),
            ])
    return suite

if __name__ == "__main__":
    import sys
    args = sys.argv[1:]
    verbosity = min(2, args.count('-v') + args.count('-vv')*2)
    unittest.TextTestRunner(verbosity=verbosity).run(suite())
